(use-modules ((srfi srfi-1)  :select (list-tabulate drop-while remove!)))
(use-modules ((srfi srfi-13) :select
              (string-contains
               string-replace substring/shared string-prefix? string-suffix?
               string-tokenize string-append/shared string-join string-index)))
(use-modules ((srfi srfi-14) :select
              (char-set-complement
               char-set string->char-set ucs-range->char-set)))
(use-modules ((ice-9 regex)  :select (match:start match:end)))

(define s+ string-append/shared)
(define s- substring/shared)

(define (drop-prefix prefix string)
  (and (string-prefix? prefix string)
       (s- string (string-length prefix))))

(define (drop-suffix suffix string)
  (and (string-suffix? suffix string)
       (s- string 0 (- (string-length string) (string-length suffix)))))

(define (proper-suffix? suffix string)
  (and (not (= (string-length suffix)
               (string-length string)))
       (string-suffix? suffix string)))

(define (substitute-substring string substring substitution)
  "\
Return a copy of STRING with all nonoverlapping occurrences of SUBSTRING
replaced by SUBSTITUTION."
  (let ((sub-len   (string-length substring))
        (subst-len (string-length substitution)))
    (let loop ((string string)
               (start 0))
      (let ((index (string-contains string substring start)))
        (if index
          (loop (string-replace string substitution index (+ index sub-len))
                (+ index subst-len))
          string)))))

(define quote/sh
  (let* ((cs (string->char-set "!\"#$&'()*;<=>?[\\]^`{|}~"))
         (cs (ucs-range->char-set 0 21 'error cs))
         (cs (ucs-range->char-set 127 256 'error cs)))
    (define (quote:sh string)
      "Return STRING, quoted for sh scripts."
      (cond
       ((string-null? string) "''")
       ((string-index string cs)
        (string-append "'" (substitute-substring string "'" "'\\''") "'"))
       (else
        string)))
    quote:sh))

(define (basename-index path)
  "\
Return one more than the index of the last slash in PATH, or zero if there
are no slashes."
  (let ((index (string-rindex path #\/)))
    (if index
      (1+ index)
      0)))

(define normalize-path
  (let ((cs (char-set-complement (char-set #\/))))
    (define (d?  string) (string=? string "."))
    (define (dd? string) (string=? string ".."))
    (define (normalize-path path)
      "\
Normalize PATH.  The result will:
* start with \"/\" iff PATH starts with \"/\";
* end with \"/\" iff PATH ends with \"/\";
* not contain any \".\" components, except possibly a single one if there are
  no other components; and
* not contain any \"..\" components, except possibly leading components in a
  relative path.
This is a stringwise operation; \"symlink/../file\" may have a different
meaning from the normalized result \"file\"."
      (let* ((index (string-index path #\000))
             (path (if index (s- path 0 index) path))
             (^/? (string-prefix? "/" path))
             (/$? (string-suffix? "/" path))
             (components (string-tokenize path cs))
             (components (remove! d? components))
             (rcomponents (reverse! components)))
        (let loop ((rcomponents rcomponents)
                   (components '()))
          (cond
           ((null? rcomponents)
            (let ((components (if ^/?
                                (drop-while dd? components)
                                components)))
              (if (null? components)
                (if ^/?
                  (if /$? "/" "/.")
                  (cond (/$? "./")
                        ((string-null? path) "")
                        (else ".")))
                (s+ (if ^/? "/" "")
                    (string-join components "/")
                    (if /$? "/" "")))))
           ((and (not (null? components))
                 (dd? (car components))
                 (not (dd? (car rcomponents))))
            (loop (cdr rcomponents) (cdr components)))
           (else
            (loop (cdr rcomponents) (cons (car rcomponents) components)))))))
    normalize-path))

(define (matcher pattern . substrings)
  (let ((regexp (make-regexp pattern))
        (table (list-tabulate (length substrings)
                              (lambda (i)
                                (cons (list-ref substrings i) (1+ i))))))
    (lambda (string)
      (let ((match (regexp-exec regexp string)))
        (and match
             (lambda (sym)
               (let* ((i (assq-ref table sym))
                      (start (match:start match i))
                      (end   (match:end   match i)))
                 (if start
                   (s- string start end)
                   ""))))))))

(define (decode:% string)
  "Decode URL-style %xx escapes in STRING."
  (let ((end (- (string-length string) 3)))
    (let loop ((decoded "")
               (start 0))
      (let ((percent (string-index encoded #\% start)))
        (cond
         ((not percent)
          (string-append decoded (s- encoded start)))
         ((> percent end)
          #f)
         (else
          (let* ((plain (s- encoded start percent))
                 (next-start (+ 3 percent))
                 (num-str (s- encoded (1+ percent) next-start))
                 (num (string->number num-str 16)))
            (and num
                 (loop (string-append decoded plain
                                      (string (integer->char num)))
                       (+ 3 percent))))))))))
